from typing import List, Dict
from pydruglogics.utils.BNetworkUtil import BNetworkUtil
import logging


class InteractionModel:
    def __init__(self, interactions_file=None, model_name='', remove_self_regulated_interactions=False,
                 remove_inputs=False,remove_outputs=False):
        """
        Initializes the InteractionModel from .sif file.
        :param interactions_file: File path to the .sif file with network interactions.
        :param model_name: Name for the model. An empty string by default.
        :param remove_self_regulated_interactions: If True, removes self-regulating interactions
        (source equals target).False by default.
        :param remove_inputs: If True, removes nodes without incoming interactions. False by default.
        :param remove_outputs: If True, removes nodes without outgoing interactions. False by default.
        """
        self._interactions: List[Dict] = []
        self._model_name = model_name

        if interactions_file is not None:
            self._load_sif_file(interactions_file)
            if remove_self_regulated_interactions:
                self._remove_self_regulated_interactions()
            self._remove_interactions(remove_inputs, remove_outputs)
            self._build_multiple_interactions()
        else:
            raise ValueError("An 'interactions_file' must be provided for initialization.")

    def _load_sif_file(self, interactions_file: str) -> None:
        """
        Loads all the lines of the .sif file and initializes the interactions.
        Each line is parsed to identify interaction types, which can include:
        - '<->' split into '<-' and '->'
        - '|-|' split into '|-' and '-|'
        - '|->' split into '->' and '|-'
        - '<-|' split into '<-' and '-|'
        :param interactions_file: File path to the .sif file with network interactions.
        :return None
        """
        try:
            file_extension = BNetworkUtil.get_file_extension(interactions_file)
            if file_extension != 'sif':
                print('New file extension used to load general model, currently not supported')
                raise IOError('ERROR: The extension needs to be .sif (other formats not yet supported)')

            self._model_name = BNetworkUtil.remove_extension(interactions_file)
            interaction_lines = BNetworkUtil.read_lines_from_file(interactions_file)

            for interaction in interaction_lines:
                if interaction:
                    if '<->' in interaction:
                        line1 = interaction.replace('<->', '<-')
                        line2 = interaction.replace('<->', '->')
                        self._interactions.extend(
                            [BNetworkUtil.parse_interaction(line1), BNetworkUtil.parse_interaction(line2)])
                    elif '|-|' in interaction:
                        line1 = interaction.replace('|-|', '|-')
                        line2 = interaction.replace('|-|', '-|')
                        self._interactions.extend(
                            [BNetworkUtil.parse_interaction(line1), BNetworkUtil.parse_interaction(line2)])
                    elif '|->' in interaction:
                        line1 = interaction.replace('|->', '->')
                        line2 = interaction.replace('|->', '|-')
                        self._interactions.extend(
                            [BNetworkUtil.parse_interaction(line1), BNetworkUtil.parse_interaction(line2)])
                    elif '<-|' in interaction:
                        line1 = interaction.replace('<-|', '<-')
                        line2 = interaction.replace('<-|', '-|')
                        self._interactions.extend(
                            [BNetworkUtil.parse_interaction(line1), BNetworkUtil.parse_interaction(line2)])
                    else:
                        self._interactions.append(BNetworkUtil.parse_interaction(interaction))

            logging.info('Interactions loaded successfully')

        except IOError as e:
            logging.error(f"Error reading the interactions file '{interactions_file}': {str(e)}")
            raise
        except Exception as e:
            logging.error(f"An unexpected error occurred while loading the file '{interactions_file}': {str(e)}")
            raise

    def _remove_interactions(self, is_input: bool = False, is_output: bool = False) -> None:
        """
        Removes interactions based on input and output criteria.
        :param is_input: Removes interactions that are only inputs. By default,  False.
        :param is_output: Removes interactions that are only outputs. By default,  False.
        :return None
        """
        interactions_size_before_trim = len(self._interactions)
        iteration_trim = 0
        if (is_input and not is_output) or (not is_input and is_output):
            logging.debug(f"Removing ({'inputs' if is_input else 'outputs'}). "
                  f"Interactions before trim: {interactions_size_before_trim}\n")
        else:
            logging.debug(f"Interactions before trim: {interactions_size_before_trim}\n")

        while True:
            iteration_trim += 1
            interactions_size_before_trim = len(self._interactions)

            for i in range(len(self._interactions) - 1, -1, -1):
                source = self._interactions[i]['source'] if is_input else None
                target = self._interactions[i]['target'] if is_output else None

                if target and self._is_not_a_source(target):
                    logging.debug(f"Removing interaction (i = {i})  (not source):  {self._interactions[i]}")
                    self._interactions.pop(i)
                if source and self._is_not_a_target(source):
                    logging.debug(f"Removing interaction (i = {i})  (not target):  {self._interactions[i]}")
                    self._interactions.pop(i)

            if interactions_size_before_trim <= len(self._interactions):
                break
        logging.debug(f"Interactions after trim ({iteration_trim} iterations): {len(self._interactions)}\n")

    def _remove_self_regulated_interactions(self) -> None:
        """
        Removes interactions that are self regulated.
        :return None
        """
        for i in range(len(self._interactions) - 1, -1, -1):
            target = self._interactions[i]['target']
            source = self._interactions[i]['source']
            if target == source:
                logging.debug(f"Removing self regulation:  {self._interactions[i]}")
                self._interactions.pop(i)

    def _build_multiple_interactions(self) -> None:
        """
        Creates interactions with multiple regulators for every single target.
        :return None
        """
        checked_targets = {}
        multiple_interaction = []
        logging.debug('Building Boolean Equations for Interactions (.sif).')
        for interaction in self._interactions:
            target = interaction['target']
            if target not in checked_targets:
                checked_targets[target] = {'activating_regulators': set(), 'inhibitory_regulators': set()}

            match interaction['arc']:
                case 1:
                    checked_targets[target]['activating_regulators'].add(interaction['source'])
                case -1:
                    checked_targets[target]['inhibitory_regulators'].add(interaction['source'])
                case _:
                    raise RuntimeError(f"ERROR: Invalid interaction detected. Source '{interaction['source']}', "
                                       f"target '{interaction['target']}' with an unsupported value "
                                       f"'{interaction['arc']}'.")

        for target, regulators in checked_targets.items():
            new_interaction = BNetworkUtil.create_interaction(target=target)
            for activating_regulator in regulators["activating_regulators"]:
                new_interaction['activating_regulators'].append(activating_regulator)
            for inhibitory_regulator in regulators["inhibitory_regulators"]:
                new_interaction['inhibitory_regulators'].append(inhibitory_regulator)
            multiple_interaction.append(new_interaction)

        sources = {interaction['source'] for interaction in self._interactions}
        for source in sources:
            if source not in checked_targets and self._is_not_a_target(source):
                interaction = BNetworkUtil.create_interaction(target=source)
                interaction['activating_regulators'].append(source)
                multiple_interaction.append(interaction)

        self._interactions = multiple_interaction

    def size(self) -> int:
        return len(self._interactions)

    def _is_not_a_source(self, node_name: str) -> bool:
        result = True
        for interaction in self._interactions:
            if node_name == interaction['source']:
                result = False
        return result

    def _is_not_a_target(self, node_name: str) -> bool:
        result = True
        for interaction in self._interactions:
            if node_name == interaction['target']:
                result = False
        return result

    def print(self) -> None:
        try:
            print(str(self))
        except Exception as e:
            logging.error(f"An error occurred while printing the Interactions: {str(e)}")
            raise

    @property
    def interactions(self) -> List[Dict]:
        return self._interactions

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def all_targets(self) -> List[str]:
        return [interaction['target'] for interaction in self.interactions]

    def get_interaction(self, index: int) -> Dict:
        return self.interactions[index]

    def get_target(self, index: int) -> str:
        return self.interactions[index]['target']

    def get_activating_regulators(self, index: int) -> List[str]:
        return self.interactions[index]['activating_regulators']

    def get_inhibitory_regulators(self, index: int) -> List[str]:
        return self.interactions[index]['inhibitory_regulators']

    @model_name.setter
    def model_name(self, model_name: str) -> None:
        self._model_name = model_name

    def __str__(self):
        multiple_interaction = ''
        for interaction in self._interactions:
            interaction_str = f"Target: {interaction['target']}"
            if interaction['activating_regulators']:
                activators_str = ', '.join(interaction['activating_regulators'])
                interaction_str += f", activating regulators: {activators_str}"

            if interaction['inhibitory_regulators']:
                inhibitors_str = ', '.join(interaction['inhibitory_regulators'])
                interaction_str += f", inhibitory regulators: {inhibitors_str}"

            multiple_interaction += interaction_str + "\n"

        return multiple_interaction



import os
import datetime
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from pydruglogics.model.BooleanModel import BooleanModel
import logging


class ModelPredictions:
    def __init__(self, boolean_models=None, perturbations=None, model_outputs=None,
                 synergy_method='bliss', model_directory=None, attractor_tool=None, attractor_type=None):
        """
        Initializes the ModelPredictions class.
        :param boolean_models: List of BooleanModel instances.
        :param perturbations: List of perturbations to apply.
        :param model_outputs: Model outputs for evaluating the predictions.
        :param observed_synergy_scores: List of observed synergy scores.
        :param synergy_method: Method to check for synergy. Possible values: 'hsa' or 'bliss'.
        :param model_directory: Directory from which to load models. (Needed only when there is no Evolution result.)
        :param attractor_tool: Tool to calculate attractors in models. (Needed only when loads models from directory.)
        :param attractor_type: Type to calculate attractors in models. (Needed only when loads models from directory.)
        """
        self._boolean_models = boolean_models or []
        self._perturbations = perturbations or []
        self._model_outputs = model_outputs
        self._synergy_method = synergy_method
        self._model_predictions = []
        self._predicted_synergy_scores = []
        self._prediction_matrix = {}

        if model_directory and not boolean_models:
            self._load_models_from_directory(directory=model_directory, attractor_tool=attractor_tool,
                                             attractor_type=attractor_type)
        if not model_directory and not boolean_models:
            raise ValueError('Please provide Boolean Models from file or list.')

    def _simulate_model_responses(self, model, perturbation):
        """
        Initializes a single perturbed Boolean model and simulates its response.
        :param model: The Boolean model to perturb.
        :param perturbation: The perturbation to apply to the model.
        :return: The perturbed model, its response, and the perturbation.
        """
        perturbed_model = model.clone()
        perturbed_model.add_perturbations(perturbation)
        logging.debug(f"Added new perturbed model: {perturbed_model.model_name}")
        perturbed_model.calculate_attractors(perturbed_model.attractor_tool, perturbed_model.attractor_type)
        global_output = perturbed_model.calculate_global_output(self._model_outputs, False)
        logging.debug(f"Adding predicted response for perturbation {perturbation}: {global_output}")
        return perturbed_model, global_output, perturbation

    def _store_result_in_matrix(self, output_matrix, model_name, perturbation, response):
        perturbation_name = self._get_perturbation_name(perturbation)

        if perturbation_name not in output_matrix:
            output_matrix[perturbation_name] = {}

        output_matrix[perturbation_name][model_name] = response

    def _get_perturbation_name(self, perturbation):
        return "-".join(drug['name'] for drug in perturbation)

    def _calculate_mean_responses(self):
        mean_values = {}
        for perturbation, model_responses in self._prediction_matrix.items():
            values = [response for response in model_responses.values() if isinstance(response, (int, float))]
            mean_values[perturbation] = np.mean(values) if values else 0
        return mean_values

    def _calculate_synergy(self):
        """
        Calculate synergy scores for perturbations that contain two drugs based on
        the chosen synergy method (HSA or Bliss).
        """
        logging.debug('\nCalculating synergies..')
        mean_responses = self._calculate_mean_responses()
        logging.info(f"\nSynergy scores ({self._synergy_method}):")
        for perturbation in self._perturbations.perturbations:
            perturbation_name = self._get_perturbation_name(perturbation)

            if '-' in perturbation_name:
                drug1, drug2 = perturbation_name.split('-')
                mean_drug1 = mean_responses.get(drug1, None)
                mean_drug2 = mean_responses.get(drug2, None)
                mean_combination = mean_responses.get(perturbation_name, None)

                if mean_drug1 is not None and mean_drug2 is not None and mean_combination is not None:
                    if self._synergy_method == 'hsa':
                        self._calculate_hsa_synergy(mean_combination, mean_drug1, mean_drug2, perturbation_name)
                    elif self._synergy_method == 'bliss':
                        self._calculate_bliss_synergy(mean_combination, mean_drug1, mean_drug2, perturbation_name)

    def _calculate_hsa_synergy(self, mean_combination, mean_drug1, mean_drug2, perturbation_name):
        min_single_drug_response = min(mean_drug1, mean_drug2)

        if mean_combination < min_single_drug_response:
            synergy_score = mean_combination - min_single_drug_response
        elif mean_combination > min_single_drug_response:
            synergy_score = mean_combination - min_single_drug_response
        else:
            synergy_score = 0

        self._predicted_synergy_scores.append((perturbation_name, synergy_score))
        logging.info(f"{perturbation_name}: {synergy_score}")

    def _calculate_bliss_synergy(self, mean_combination, mean_drug1, mean_drug2, perturbation_name):
        drug1_response = ((mean_drug1 - self._model_outputs.min_output) /
                            (self._model_outputs.max_output - self._model_outputs.min_output))
        drug2_response = ((mean_drug2 - self._model_outputs.min_output) /
                          (self._model_outputs.max_output - self._model_outputs.min_output))
        expected_combination_response = 1.0 * drug1_response * drug2_response
        combination_response = ((mean_combination - self._model_outputs.min_output) /
                                (self._model_outputs.max_output - self._model_outputs.min_output))
        synergy_score = combination_response - expected_combination_response

        self._predicted_synergy_scores.append((perturbation_name, synergy_score))
        logging.info(f"{perturbation_name}: {synergy_score}")

    def _load_models_from_directory(self, directory, attractor_tool, attractor_type):
        """
        Loads all .bnet files from the given directory into Boolean Models with attractors and global outputs.
        :param directory: The path to the directory containing '.bnet' files.
        :param attractor_tool: The tool used for attractor calculation. Possible values : 'mpbn', 'pyboolnet'.
        :param attractor_type: The type of attractor to be calculated. Possible values: 'stable_states', 'trapspaces'.
        :return: None
        """
        try:
            for filename in os.listdir(directory):
                if filename.endswith('.bnet'):
                    file_path = os.path.join(directory, filename)
                    try:
                        model = BooleanModel(file=file_path, attractor_tool=attractor_tool,
                                             attractor_type=attractor_type)
                        model.calculate_attractors(attractor_tool, attractor_type)
                        model.calculate_global_output(self._model_outputs, False)
                        self._boolean_models.append(model)
                        logging.debug(f"Model loaded from {file_path}")
                    except Exception as e:
                        print(f"Failed to load model from {file_path}: {str(e)}")
                        raise
        except Exception as e:
            logging.error(f"Error loading models from directory {directory}: {str(e)}")
            raise

    def run_simulations(self, parallel=True, cores=4):
        """
        Runs simulations on the Boolean Models with the perturbations, either in parallel or serially.
        :param parallel: Whether to run the simulations in parallel By default, True.
        :param cores: The number of CPU cores to use for parallel execution By default, 4.
        :return: None
        """
        self._model_predictions = []
        self._prediction_matrix = {}

        if parallel:
            logging.debug('Running simulations in parallel.')
            results = Parallel(n_jobs=cores, backend='loky')(
                delayed(self._simulate_model_responses)(model, perturbation)
                for model in self._boolean_models
                for perturbation in self._perturbations.perturbations
            )

            for model, global_output, perturbation in results:
                self._store_result_in_matrix(self._prediction_matrix, model.model_name, perturbation, global_output)
                self._model_predictions.append((model.model_name, global_output, perturbation))
        else:
            logging.debug('Running simulations serially.')
            for model in self._boolean_models:
                for perturbation in self._perturbations.perturbations:
                    model, global_output, perturbation = self._simulate_model_responses(model, perturbation)
                    self._store_result_in_matrix(self._prediction_matrix, model.model_name, perturbation, global_output)
                    self._model_predictions.append((model.model_name, global_output, perturbation))

        self.get_prediction_matrix()
        self._calculate_synergy()

    def get_prediction_matrix(self):
        filtered_matrix = {k: v for k, v in self._prediction_matrix.items() if k.count('-') == 1}
        response_matrix_df = pd.DataFrame.from_dict(filtered_matrix, orient='index').fillna('NA')

        pd.set_option('display.max_columns', None)
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_colwidth', None)
        logging.debug("\nResponse Matrix:")
        logging.debug(response_matrix_df)

    def save_to_file_predictions(self, base_folder='./results/predictions'):
        try:
            time_now = datetime.datetime.now()
            current_date = time_now.strftime('%Y_%m_%d')
            current_time = time_now.strftime('%H%M')

            subfolder = os.path.join(base_folder, f"predictions_{current_date}_{current_time}")
            if not os.path.exists(subfolder):
                os.makedirs(subfolder)

            with open(os.path.join(subfolder, "model_scores.tab"), "w") as file:
                filtered_matrix = {k: v for k, v in self._prediction_matrix.items() if k.count('-') == 1}
                response_matrix_df = pd.DataFrame.from_dict(filtered_matrix, orient='index').fillna('NA')

                file.write("# Perturbed scores\n")
                response_matrix_df.to_csv(file, sep='\t', mode='w')

                file.write("\n# Unperturbed scores\n")
                for model in self._boolean_models:
                    file.write(f"{model.model_name}\t{model.global_output}\n")

            with open(os.path.join(subfolder, f"synergies_{self._synergy_method}.tab"), "w") as file:
                file.write(f"# Synergies ({self._synergy_method})\n")
                file.write("perturbation_name\tsynergy_score\n")
                for perturbation, score in self._predicted_synergy_scores:
                    if perturbation.count('-') == 1:
                        file.write(f"{perturbation}\t{score}\n")

            logging.info(f"Predictions saved to {subfolder}")

        except IOError as e:
            logging.error(f"File I/O error while saving predictions: {str(e)}")
            raise
        except Exception as e:
            logging.error(f"Error saving predictions to file: {str(e)}")
            raise

    @property
    def predicted_synergy_scores(self):
        return self._predicted_synergy_scores



import logging
from typing import List, Dict
from pydruglogics.utils.BNetworkUtil import BNetworkUtil


class ModelOutputs:
    def __init__(self, input_file: str = None, input_dictionary: Dict[str, float] = None):
        self._model_outputs: Dict[str, float] = {}
        if input_file is not None:
            self._load_model_outputs_file(input_file)
        elif input_dictionary is not None:
            self._load_model_outputs_dict(input_dictionary)
        else:
            raise ValueError('Provide either a file or a dictionary for initialization.')

        self._min_output = self._calculate_min_output()
        self._max_output = self._calculate_max_output()


    def _calculate_max_output(self) -> float:
        return sum(max(weight, 0) for weight in self._model_outputs.values())

    def _calculate_min_output(self) -> float:
        return sum(min(weight, 0) for weight in self._model_outputs.values())

    def _load_model_outputs_file(self, file: str):
        try:
            lines = BNetworkUtil.read_lines_from_file(file, True)
            for line in lines:
                node_name, weight = map(str.strip, line.split("\t"))
                self._model_outputs[node_name] = float(weight)

            logging.info(f"Model outputs loaded from file: {file}.")
        except IOError as e:
            logging.error(f"File read error: {str(e)}")
            raise
        except Exception as e:
            logging.error(f"Error while loading model outputs from file: {str(e)}")
            raise

    def _load_model_outputs_dict(self, model_outputs_dict: Dict[str, float]):
        if not isinstance(model_outputs_dict, dict):
            raise TypeError("model_outputs_dict must be a dictionary [text, number].")

        if not model_outputs_dict:
            logging.warning("model_outputs_dict is empty. No model outputs loaded.")

        self._model_outputs = model_outputs_dict
        logging.info("Model outputs are initialized from dictionary.")

    def get_model_output(self, node_name: str) -> float:
        return self._model_outputs.get(node_name, 0.0)

    def size(self) -> int:
        return len(self._model_outputs)

    def print(self) -> None:
        try:
            print(str(self))
        except Exception as e:
            logging.error(f"An error occurred while printing ModelOutputs: {str(e)}")
            raise

    @property
    def node_names(self) -> List[str]:
        return list(self._model_outputs.keys())

    @property
    def model_outputs(self) -> Dict[str, float]:
        return self._model_outputs

    @property
    def min_output(self) -> float:
        return self._min_output

    @property
    def max_output(self) -> float:
        return self._max_output

    def __str__(self) -> str:
        return "\n".join(f"Model output: {node_name}, weight: {weight}"
                         for node_name, weight in self._model_outputs.items())



import logging
import sys


class Logger:

    @classmethod
    def set_logger(cls, level=logging.INFO, external_lib_off=True):
        logger = logging.getLogger()

        if logger.hasHandlers():
            logger.handlers.clear()

        logger.setLevel(level)

        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.DEBUG)
        stdout_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        stdout_handler.setFormatter(stdout_formatter)

        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.setLevel(logging.WARNING)
        stderr_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        stderr_handler.setFormatter(stderr_formatter)

        logger.addHandler(stdout_handler)
        logger.addHandler(stderr_handler)

        if external_lib_off:
            for module_name, module in logging.Logger.manager.loggerDict.items():
                if isinstance(module, logging.Logger):
                    module.setLevel(logging.INFO)

Logger.set_logger(level=logging.INFO)



from abc import ABC, abstractmethod
from pydruglogics.model import BooleanModel


class BooleanModelOptimizer(ABC):
    @abstractmethod
    def run(self) -> [BooleanModel]:
        return []

    @abstractmethod
    def save_to_file_models(self, path):
        pass



import itertools
from typing import List, Dict, Union
import logging


class Perturbation:
    def __init__(self, drug_data: List[List[Union[str, None]]] = None, perturbation_data: List[List[str]] = None):
        self._drug_panel = []
        self._perturbations = []
        self._drug_perturbations = []

        if drug_data is not None:
            self._load_drug_panel_from_data(drug_data)
        else:
            raise ValueError('Please provide drug data.')

        if perturbation_data is not None:
            self._load_perturbations_from_data(perturbation_data)
        else:
            self._init_perturbations_from_drugpanel()

    def _load_drug_panel_from_data(self, drug_data: List[List[Union[str, None]]]) -> None:
        for drug in drug_data:
            if len(drug) < 2:
                raise ValueError("Each drug entry must contain at least 'name' and 'targets'.")

            name, targets, effect = drug[0], drug[1], drug[2] if len(drug) > 2 else 'inhibits'
            self._drug_panel.append({
                'name': name,
                'targets': targets.split(',') if isinstance(targets, str) else targets,
                'effect': effect
            })
        logging.info("Drug panel data initialized from list.")

    def _load_perturbations_from_data(self, perturbation_data: List[List[str]]) -> None:
        if not perturbation_data:
            logging.warning("Perturbation data is empty. Generating perturbations from the drug panel.")
            return

        self._drug_perturbations = [entry for entry in perturbation_data if entry]
        if len(self._drug_perturbations) < len(perturbation_data):
            logging.warning("Some perturbation entries were empty and have been ignored.")

        self._init_drug_perturbations()

    def _init_drug_perturbations(self) -> None:
        drug_name_in_panel = {drug['name']: drug for drug in self._drug_panel}
        perturbed_drugs = []

        for combination_in_perturbations in self._drug_perturbations:
            combination_in_panel = [drug_name_in_panel.get(name) for name in combination_in_perturbations if name in drug_name_in_panel]
            if len(combination_in_panel) == len(combination_in_perturbations):
                perturbed_drugs.append(combination_in_panel)
            else:
                logging.warning('Some drugs in the perturbation were not found in the drug panel.')

        self._perturbations = perturbed_drugs
        logging.info("Drug perturbations initialized.")

    def _init_perturbations_from_drugpanel(self):
        self._perturbations = [
            list(combination)
            for number_of_combination in range(1, 3)
            for combination in itertools.combinations(self._drug_panel, number_of_combination)
        ]
        logging.info("Perturbations generated from drug panel.")

    @property
    def drugs(self) -> List[Dict[str, str]]:
        return self._drug_panel

    @property
    def perturbations(self) -> List[List[Dict[str, str]]]:
        return self._perturbations

    @property
    def drug_names(self) -> List[str]:
        return [drug['name'] for drug in self._drug_panel]

    @property
    def drug_effects(self) -> List[str]:
        return [drug['effect'] for drug in self._drug_panel]

    @property
    def drug_targets(self) -> List[List[str]]:
        return [drug['targets'] for drug in self._drug_panel]

    def print(self) -> None:
        try:
            print(str(self))
        except Exception as e:
            print(f"An error occurred while printing Perturbation: {e}")

    def __str__(self) -> str:
        if not self._perturbations:
            return 'No perturbations available.'
        perturbations_str = []
        for perturbation in self._perturbations:
            combo_str = ', '.join([f"{drug['name']} (targets: {', '.join(drug['targets'])})" for drug in perturbation])
            perturbations_str.append(f"[{combo_str}]")
        return '\n'.join(perturbations_str)



import logging
import math
import os
from datetime import datetime
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d
from scipy.stats import norm
from sklearn.metrics import precision_recall_curve, auc
from pydruglogics.model.ModelPredictions import ModelPredictions
from pydruglogics.utils.PlotUtil import PlotUtil
from typing import List, Tuple, Any

def _normalize_synergy_scores(calibrated_synergy_scores: List[Tuple[str, float]],
                              prolif_synergy_scores: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
    """
    Normalize synergy scores based on calibrated and proliferative synergy scores.
    :param calibrated_synergy_scores: List of tuples (perturbation, calibrated synergy score).
    :param prolif_synergy_scores: List of tuples (perturbation, proliferative synergy score).
    :return: List of tuples (perturbation, normalized synergy score).
    """
    normalized_synergy_scores = []
    for (perturbation, ss_score), (_, prolif_score) in zip(calibrated_synergy_scores, prolif_synergy_scores):
        normalized_synergy_score = math.exp(ss_score - prolif_score)
        normalized_synergy_scores.append((perturbation, normalized_synergy_score))

    return normalized_synergy_scores

def _bootstrap_resample(labels: np.ndarray, predictions: np.ndarray, boot_n: int) -> List[Tuple[np.ndarray, np.ndarray]]:
    """
    Resample data for bootstrapping.
    :param labels: Array of observed binary labels.
    :param predictions: Array of predicted scores.
    :param boot_n: Number of bootstrap resampling iterations.
    :return: List of tuples containing resampled labels and predictions.
    """
    resampled_model_preds = []
    for _ in range(boot_n):
        rnd = np.random.choice(len(labels), size=len(labels), replace=True)
        resampled_labels = labels[rnd]
        resampled_predictions = predictions[rnd]
        resampled_model_preds.append((resampled_labels, resampled_predictions))
    return resampled_model_preds

def _calculate_pr_with_ci(observed: np.ndarray, preds: np.ndarray, boot_n: int,
                          confidence_level: float, with_seeds: bool, seeds: int) -> Tuple[pd.DataFrame, float, dict]:
    """
    Calculate Precision-Recall curve with confidence intervals.
    :param observed: Array of observed binary labels.
    :param preds: Array of predicted scores.
    :param boot_n: Number of bootstrap resampling iterations.
    :param confidence_level: Confidence level for calculating the confidence intervals (e.g., 0.9 for 90% CI).
    :param with_seeds: Whether to use a fixed seed for reproducibility of the bootstrap sampling.
    :param seeds: Seed value for random number generation to ensure reproducibility.
    :return: A tuple containing a DataFrame with 'recall', 'precision', 'low_precision', and 'high_precision'
    columns for confidence intervals the AUC-PR.
    """
    if with_seeds:
        np.random.seed(seeds)

    precision_orig, recall_orig, _ = precision_recall_curve(observed, preds)
    auc_pr = auc(recall_orig, precision_orig)
    pr_df = pd.DataFrame({'recall': recall_orig, 'precision': precision_orig})

    resampled_data = _bootstrap_resample(observed, preds, boot_n=boot_n)
    precision_matrix = []

    for resampled_observed, resampled_predicted in resampled_data:
        precision_boot, recall_boot, _ = precision_recall_curve(resampled_observed, resampled_predicted)
        const_interp_pr = interp1d(recall_boot, precision_boot, kind='previous', bounds_error=False,
                                   fill_value=(precision_boot[0], precision_boot[-1]))
        aligned_precisions = const_interp_pr(recall_orig)
        precision_matrix.append(aligned_precisions)

    precision_matrix = np.array(precision_matrix)

    alpha = 1 - confidence_level
    low_precision = np.percentile(precision_matrix, alpha / 2 * 100, axis=0)
    high_precision = np.percentile(precision_matrix, (1 - alpha / 2) * 100, axis=0)

    pr_df['low_precision'] = low_precision
    pr_df['high_precision'] = high_precision

    sample_mean = np.mean(precision_orig)
    standard_deviation = np.std(precision_orig)
    standard_err = standard_deviation/np.sqrt(len(precision_orig))
    critical_value = norm.ppf(1-alpha/ 2)
    margin_of_err = critical_value * standard_err
    ci_lower = float(sample_mean - margin_of_err)
    ci_upper = float(sample_mean + margin_of_err)

    sampl_with_ci_results= {
        "Point Estimate (Mean)": sample_mean,
        "Standard Deviation": standard_deviation,
        "Standard Error": standard_err,
        "Confidence Interval": (ci_lower, ci_upper),
        "Confidence Level": f"{confidence_level * 100}%",
        "Critical Value": critical_value,
        "Margin of Error": margin_of_err,
        "Sample Size": len(precision_orig)
    }

    return pr_df, auc_pr, sampl_with_ci_results


def _create_result_base_folder(main_folder: str, category_folder: str, prefix: str) -> str:

    base_path = os.path.join(main_folder, category_folder)
    os.makedirs(base_path, exist_ok=True)
    timestamp = datetime.now().strftime('%Y_%m_%d_%H%M')
    result_folder = os.path.join(base_path, f"{prefix}_{timestamp}")
    os.makedirs(result_folder, exist_ok=True)

    return result_folder


def _save_sampling_results(synergy_scores: List[List[Tuple[str, float]]], base_folder: str, synergy_method: str,
                           summary_metrics: dict):
    current_time = datetime.now().strftime('%Y/%m/%d %H:%M')
    summary_path = os.path.join(base_folder, "sampling_with_ci.tab")

    with open(summary_path, 'w') as file:
        file.write(f"Date: {current_time.split()[0]}, Time: {current_time.split()[1]}\n")
        file.write("Sampling Results\n")
        for key, value in summary_metrics.items():
            formatted_value = f"({value[0]:.6f}, {value[1]:.6f})" if isinstance(value, tuple) else f"{value:.6f}" if isinstance(value, float) else str(value)
            file.write(f"{key}: {formatted_value}\n")

    for idx, sample in enumerate(synergy_scores, start=1):
        file_path = os.path.join(base_folder, f"synergy_scores_sample_{idx}.tab")
        with open(file_path, 'w') as file:
            file.write(f"# Date: {current_time.split()[0]}, Time: {current_time.split()[1]}\n")
            file.write(f"# Synergies ({synergy_method})\n")
            file.write("perturbation_name\tsynergy_score\n")
            for perturbation, score in sample:
                file.write(f"{perturbation}\t{score}\n")

    logging.info(f"Sampling results saved to {base_folder}")


def _save_compare_results(predicted_synergy_scores_list: List[List[Tuple[str, float]]], labels: List[str],
                          base_folder: str, synergy_method: str):
    current_time = datetime.now().strftime('%Y/%m/%d %H:%M')

    for scores, label in zip(predicted_synergy_scores_list, labels):
        underscore_label = label.replace(" ", "_").lower()
        file_path = os.path.join(base_folder, f"predicted_synergy_scores_{underscore_label}.tab")

        with open(file_path, 'w') as file:
            file.write(f"# Date: {current_time.split()[0]}, Time: {current_time.split()[1]}\n")
            file.write(f"# Synergies ({synergy_method})\n")
            file.write("perturbation_name\tsynergy_score\n")
            for perturbation, score in scores:
                file.write(f"{perturbation}\t{score}\n")

    logging.info(f"Comparison results saved to {base_folder}")

def sampling_with_ci(boolean_models: List, observed_synergy_scores: List[str], model_outputs: Any,
                     perturbations: Any, synergy_method: str = 'bliss', repeat_time: int = 10, sub_ratio: float = 0.8,
                     boot_n: int = 1000, confidence_level: float = 0.9, plot: bool = True, plot_discrete: bool = False,
                     save_result: bool = True, with_seeds: bool = True, seeds: int = 42) -> None:
    """
    Performs sampling with confidence interval calculation and plot the PR curve.
    :param boolean_models: List of BooleanModel instances.
    :param observed_synergy_scores: List of observed synergy scores.
    :param model_outputs: Model outputs for evaluation.
    :param perturbations: List of perturbations to apply to the models.
    :param synergy_method: Method to check for synergy ('hsa' or 'bliss').
    :param repeat_time: Number of times to repeat sampling.
    :param sub_ratio: Proportion of models to sample in each iteration.
    :param boot_n: Number of bootstrap resampling iterations for confidence intervals.
    :param confidence_level: Confidence level for confidence interval calculations.
    :param plot: Whether to display the PR curve.
    :param plot_discrete: Whether to plot discrete points on the PR curve.
    :param save_result: Whether to save the results to a .tab file.
    :param with_seeds: Whether to use a fixed seed for reproducibility.
    :param seeds: Seed value for random number generation to ensure reproducibility.
    :return: None
    """

    base_folder = _create_result_base_folder('results', 'sampling', 'sampling')
    num_models = len(boolean_models)
    sample_size = int(sub_ratio * num_models)
    predicted_synergy_scores_list = []

    for i in range(repeat_time):
        if with_seeds:
            np.random.seed(seeds + i)
        sampled_models = np.random.choice(boolean_models, size=sample_size, replace=False).tolist()

        model_predictions = ModelPredictions(
            boolean_models=sampled_models,
            perturbations=perturbations,
            model_outputs=model_outputs,
            synergy_method=synergy_method
        )
        model_predictions.run_simulations(parallel=True)
        predicted_synergy_scores_list.append(model_predictions.predicted_synergy_scores)

    all_predictions = []
    all_observed = []

    for pred_synergy_scores in predicted_synergy_scores_list:
        df = pd.DataFrame(pred_synergy_scores, columns=['perturbation', 'synergy_score'])
        df['observed'] = df['perturbation'].apply(lambda pert: 1 if pert in observed_synergy_scores else 0)
        df['synergy_score'] *= -1
        all_predictions.extend(df['synergy_score'].values)
        all_observed.extend(df['observed'].values)

    pr_df, auc_pr, summary_metrics = _calculate_pr_with_ci(np.array(all_observed), np.array(all_predictions),
                                                           boot_n, confidence_level, with_seeds, seeds)

    if save_result:
        _save_sampling_results(synergy_scores=predicted_synergy_scores_list, base_folder=base_folder,
                               synergy_method=synergy_method, summary_metrics=summary_metrics)

    if plot:
        PlotUtil.plot_pr_curve_with_ci(pr_df, auc_pr, boot_n=boot_n, plot_discrete=plot_discrete)


def compare_two_simulations(boolean_models1: List, boolean_models2: List, observed_synergy_scores: List[str],
                            model_outputs: Any, perturbations: Any, synergy_method: str = 'bliss',
                            label1: str = 'Models 1', label2: str = 'Models 2', normalized: bool = True,
                            plot: bool = True, save_result: bool = True) -> None:
    """
    Compares ROC and PR curves for two sets of evolution results.
    By default, normalization of the first result is applied.
    :param boolean_models1: List of the best Boolean Models for the first simulation set.
    :param boolean_models2: List of the best Boolean Models for the second simulation set.
    :param observed_synergy_scores: List of observed synergy scores for comparison.
    :param model_outputs: Model outputs for evaluation.
    :param perturbations: List of perturbations to apply to the models.
    :param synergy_method: Method to check for synergy ('hsa' or 'bliss').
    :param label1: Label for the first simulation result.
    :param label2: Label for the second simulation result.
    :param normalized: Whether to normalize the first result.
    :param plot: Whether to display the ROC and PR curves.
    :param save_result: Whether to save the results.
    :return: None
    """
    base_folder = _create_result_base_folder('results', 'comparison', 'comparison')

    predicted_synergy_scores_list = []
    labels = [label1, label2]

    model_predictions1 = ModelPredictions(
        boolean_models=boolean_models1,
        perturbations=perturbations,
        model_outputs=model_outputs,
        synergy_method=synergy_method
    )
    model_predictions1.run_simulations(parallel=True)
    predicted_synergy_scores1 = model_predictions1.predicted_synergy_scores
    predicted_synergy_scores_list.append(predicted_synergy_scores1)

    model_predictions2 = ModelPredictions(
        boolean_models=boolean_models2,
        perturbations=perturbations,
        model_outputs=model_outputs,
        synergy_method=synergy_method
    )
    model_predictions2.run_simulations(parallel=True)
    predicted_synergy_scores2 = model_predictions2.predicted_synergy_scores
    predicted_synergy_scores_list.append(predicted_synergy_scores2)

    if normalized:
        normalized_synergy_scores = _normalize_synergy_scores(predicted_synergy_scores1, predicted_synergy_scores2)
        predicted_synergy_scores_list.append(normalized_synergy_scores)
        labels.append('Calibrated (Normalized)')

    if save_result:
        _save_compare_results(predicted_synergy_scores_list, labels, base_folder=base_folder, synergy_method=synergy_method)

    if plot:
        PlotUtil.plot_roc_and_pr_curve(predicted_synergy_scores_list, observed_synergy_scores, synergy_method, labels)



import logging

from pydruglogics.input.Perturbations import Perturbation
from pydruglogics.input.TrainingData import TrainingData
from pydruglogics.input.ModelOutputs import ModelOutputs
from pydruglogics.model.BooleanModel import BooleanModel
from pydruglogics.model.Evolution import Evolution
from pydruglogics.model.InteractionModel import InteractionModel
from pydruglogics.model.ModelPredictions import ModelPredictions
from pydruglogics.statistics.Statistics import compare_two_simulations, sampling_with_ci
from pydruglogics.execution.Executor import execute, train
from pydruglogics.utils.Logger import Logger
from pydruglogics.utils.PlotUtil import PlotUtil

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


__version__ = "0.1.2"

__all__ = [
    "Perturbation",
    "TrainingData",
    "ModelOutputs",
    "BooleanModel",
    "Evolution",
    "InteractionModel",
    "ModelPredictions",
    "compare_two_simulations",
    "sampling_with_ci",
    "execute",
    "train",
    "Logger",
    "PlotUtil"
]


from typing import List, Dict, Union, Tuple
from pydruglogics.utils.BNetworkUtil import BNetworkUtil
import logging

class TrainingData:
    def __init__(self, input_file: str = None, observations: List[Tuple[List[str], float]] = None):
        self._observations = []
        if input_file is not None:
            self._load_from_file(input_file)
        elif observations is not None:
            self._load_from_observations_list(observations)
        else:
            raise ValueError('Please provide a dictionary or a file.')

    def _load_from_file(self, file: str) -> None:
        try:
            lines = BNetworkUtil.read_lines_from_file(file)
        except IOError as e:
            logging.error(f"File read error: {str(e)}")
            raise

        line_index = 0
        condition, response = None, None

        while line_index < len(lines):
            line = lines[line_index].strip().lower()

            if line.startswith('condition'):
                condition = lines[line_index + 1].split("\t")
                line_index += 1
            elif line.startswith('response'):
                response = lines[line_index + 1].split("\t")
                value = response[0].split(":")[1] if 'globaloutput' in response[0] else None

                if 'globaloutput' in response[0]:
                    value = response[0].split(":")[1]

                    if not BNetworkUtil.is_numeric_string(value):
                        raise ValueError(f"Response: {response} has a non-numeric value: {value}")
                    if not (-1.0 <= float(value) <= 1.0):
                        raise ValueError(f"Response has globaloutput outside the [-1,1] range: {value}")

                line_index += 1

            elif line.startswith('weight'):
                weight = float(line.split(':')[1])

                if condition is None or response is None:
                    raise ValueError("Missing condition or response data before the weight entry.")

                self._add_observation(condition, response, weight)

            line_index += 1

        logging.info(f"Training data loaded from file: {file}.")

    def _load_from_observations_list(self, observations: List[Tuple[List[str], float]]) -> None:
        for observation in observations:
            response, weight = observation
            self._add_observation(['-'], response, weight)

        logging.info('Training data initialized from list.')

    def _add_observation(self, condition: List[str], response: List[str], weight: float) -> None:
        if 'globaloutput' in response[0]:
            value = response[0].split(":")[1]
            if not BNetworkUtil.is_numeric_string(value):
                raise ValueError(f"Response: {response} has a non-numeric value: {value}")
            if not (-1.0 <= float(value) <= 1.0):
                raise ValueError(f"Response has globaloutput outside the [-1,1] range: {value}")

        self._observations.append({
            'condition': condition,
            'response': response,
            'weight': weight
        })

    def print(self) -> None:
        try:
            print(str(self))
        except Exception as e:
            logging.error(f"Error while printing TrainingData: {str(e)}")
            raise

    @property
    def weight_sum(self) -> float:
        return sum(observation['weight'] for observation in self._observations)

    def size(self) -> int:
        return len(self._observations)

    @property
    def observations(self) -> List[Dict[str, Union[str, float]]]:
        return self._observations

    @property
    def responses(self) -> List[str]:
        return [item for sublist in (obs['response'] for obs in self._observations) for item in sublist]

    @property
    def response(self) -> List[str]:
        return self._observations[0]['response'] if self._observations else []

    @property
    def weights(self) -> List[float]:
        return [obs['weight'] for obs in self._observations]

    def __str__(self) -> str:
        if not self._observations:
            return "No observations available."
        observations_str = []
        for observation in self._observations:
            observations_str.append(
                f"Observation:\nCondition: {', '.join(observation['condition'])}\n"
                f"Response: {', '.join(observation['response'])}\n"
                f"Weight: {observation['weight']}\n"
            )
        return "\n".join(observations_str)



import copy
import os
import logging
import numpy as np
import mpbn
from pyboolnet.file_exchange import bnet2primes
from pyboolnet.trap_spaces import compute_trap_spaces, compute_steady_states
from pydruglogics.utils.BNetworkUtil import BNetworkUtil


class BooleanModel:
    def __init__(self, model=None,  file=None, attractor_tool='mpbn', attractor_type='stable_states',
                 mutation_type='topology', model_name='', cloned_equations=None, cloned_binary_equations=None):
        """
        Initializes the BooleanModel instance.
        :param model: An InteractionModel instance.
        :param file: The path to the file containing Boolean Equations in '.bnet' format.
        :param attractor_tool: The tool to be used for attractor calculation. Possible values: 'mpbn', 'pyboolnet'.
        :param attractor_type: The type of the attractor calculation. Possible values: 'stable_states', 'trapspaces'.
        :param mutation_type: The type of mutation to be performed. Possible values: 'topology', 'binary', 'balanced'.
        :param model_name: Name for the model.
        :param cloned_equations: Boolean Equations representing the model's interactions. (only for cloning)
        :param cloned_binary_equations: A list representing the Mutate Boolean Model in binary representation.
        (only for cloning)
        """
        self._model_name = model_name
        self._boolean_equations = []
        self._binary_boolean_equations = np.array([])
        self._attractors = {}
        self._attractor_tool = attractor_tool
        self._attractor_type = attractor_type
        self._mutation_type = mutation_type
        self._global_output = 0.0
        self._is_bnet_file = False
        self._bnet_equations = ''

        if attractor_tool not in ['mpbn', 'pyboolnet'] or attractor_type not in ['stable_states', 'trapspaces']:
            raise ValueError("Invalid attractor tool or type. Use 'mpbn' or 'pyboolnet' for attractor_tool, and "
                             "'stable_states' or 'trapspaces' for attractor_type.")

        if mutation_type not in ['topology', 'mixed', 'balanced']:
            raise ValueError("Invalid mutation type. Use 'topology' or 'mixed' or 'balanced' for mutation_type.")

        try:
            if model is not None: # init from InteractionModel instance
                self._init_from_model(model)
                self.to_binary(self._mutation_type)
            elif file is not None:
                self._init_from_bnet_file(file) # init from .bnet file
                self.to_binary(self._mutation_type)
            elif cloned_binary_equations is not None:
                self._init_from_equations(cloned_equations, cloned_binary_equations) # for cloning
            else:
                raise ValueError('Initialization failed: Please provide a model or a file for the Boolean Model.')
        except Exception as e:
            logging.error(f"Error occurred during the initialization: {str(e)}")
            raise

    def _init_from_model(self, model) -> None:
        """
        Initialize the BooleanModel from an InteractionModel instance.
        :param model: The InteractionModel instance containing interactions.
        """
        self._model_name = model.model_name

        self._boolean_equations.extend(self._create_equation_from_interaction(model, i) for i in range(model.size()))

        logging.info('Boolean Model from Interaction Model is created.')

    def _init_from_bnet_file(self, file: str) -> None:
        """
        Initialize the BooleanModel from a '.bnet' file.
        :param file: The directory of the '.bnet' file.
        :return: None
        """
        logging.debug(f"Loading Boolean Model from file: {file}")
        try:
            with open(file, 'r') as model_file:
                lines = model_file.readlines()

            if BNetworkUtil.get_file_extension(file) != 'bnet':
                raise ValueError('The file extension has to be .bnet format.')

            self._boolean_equations = []
            self._model_name = os.path.splitext(os.path.basename(file))[0]

            for line in lines:
                if line.strip().startswith('#') or line.strip().startswith('targets') or not line.strip():
                    continue
                equation = line.strip()
                parsed_equation_bnet = BNetworkUtil.create_equation_from_bnet(equation)
                self._bnet_equations += f"{equation}\n"
                self._boolean_equations.append(parsed_equation_bnet)
                self._is_bnet_file = True

            logging.info('Boolean Model from .bnet file is created.')

        except IOError as e:
            logging.error(f"Error reading file: {str(e)}")
            raise

    def _init_from_equations(self, equations, binary_equations):
        """
        Init for cloning.
        :param equations: Boolean Equations for cloning.
        :param binary_equations: Binary Equations for cloning.
        :return: None
        """
        self._boolean_equations = equations
        self._binary_boolean_equations = np.array(binary_equations)

    def _get_equations_in_bnet_format(self):
        if self._is_bnet_file:
            self._is_bnet_file = False
            return self._bnet_equations
        return BNetworkUtil.to_bnet_format(self._boolean_equations)

    def _calculate_attractors_mpbn(self, attractor_type):
        """
        Calculate attractors using MPBN based on attractor_type.
        :param attractor_type: Attractor type for calculation  ('stable_states' or 'trapspaces').
        :return: None
        """
        equations_bnet_dict = BNetworkUtil.bnet_string_to_dict(self._get_equations_in_bnet_format())
        boolean_network_mpbn = mpbn.MPBooleanNetwork(equations_bnet_dict)

        self._attractors = list(boolean_network_mpbn.fixedpoints() if attractor_type == 'stable_states'
                                else boolean_network_mpbn.attractors())

        logging.debug(f"MPBN found {len(self._attractors)} attractor(s):\n{self._attractors}")

    def _calculate_attractors_pyboolnet(self, attractor_type):
        """
        Calculate attractors using PyBoolNet based on attractor_type.
        :param attractor_type: Attractor type for calculation ('stable_states' or 'trapspaces').
        :return: None
        """
        primes = bnet2primes(self._get_equations_in_bnet_format())

        if attractor_type == 'trapspaces':
            self._attractors = compute_trap_spaces(primes)
            all_node_in_model = {target for target, _, _, _ in self._boolean_equations}

            self._attractors = [
                {**{node: '*' for node in all_node_in_model}, **attractor}
                if len(attractor) < len(all_node_in_model) else attractor
                for attractor in self._attractors
            ]

        elif attractor_type == 'stable_states':
            self._attractors = compute_steady_states(primes)

        logging.debug(f"PyBoolNet found {len(self._attractors)} attractor(s):\n{self._attractors}")

    def _create_equation_from_interaction(self, interaction, interaction_index):
        """
        Create a Boolean equation from an InteractionModel.
        :param interaction: An interaction of the InteractionModel.
        :param interaction_index: Index of an interaction.
        :return: Equation dictionary with target, regulators, operators.
        """
        activating_regulators = {}
        inhibitory_regulators = {}

        target = interaction.get_target(interaction_index)
        tmp_activating_regulators = interaction.get_activating_regulators(interaction_index)
        tmp_inhibitory_regulators = interaction.get_inhibitory_regulators(interaction_index)
        link = '' if not tmp_activating_regulators or not tmp_inhibitory_regulators else '&'

        for i, regulator in enumerate(tmp_activating_regulators):
            activating_regulators[regulator] = 1

        for i, regulator in enumerate(tmp_inhibitory_regulators):
            inhibitory_regulators[regulator] = 1

        return (target, activating_regulators, inhibitory_regulators, link,)

    def _perturb_nodes(self, node_names, effect):
        """
        Apply a perturbation to the given nodes by modifying their Boolean equations.
        :param node_names: A list of node names to be perturbed.
        :param effect: The type of perturbation effect ('inhibits' sets the value to 0, others set it to 1).
        :return: None
        """
        value = 0 if effect == 'inhibits' else 1

        for node in node_names:
            for i, equation in enumerate(self._boolean_equations):
                target, _, _, _= equation
                if node == target:
                    new_equation = (node, {str(value): 1}, {}, '')
                    self._boolean_equations[i] = new_equation
                    break

    def calculate_attractors(self, attractor_tool, attractor_type) -> None:
        """
        Calculate attractors based on the chosen attractor_tool and attractor_type.
        :param attractor_tool: Tool for attractor calculation. Possible values: 'mpbn', 'pyboolnet'.
        :param attractor_type: Type of attractor. Possible values: 'stable_states', 'trapspaces'.
        :return: None
        """
        tool_methods = {
            'mpbn': self._calculate_attractors_mpbn,
            'pyboolnet': self._calculate_attractors_pyboolnet
        }

        if attractor_tool in tool_methods:
            tool_methods[attractor_tool](attractor_type)
        else:
            raise ValueError("Please provide a valid attractor tool and type. Valid tools: 'mpbn', 'pyboolnet'. "
                             "Valid types: 'stable_states', 'trapspaces'.")


    def calculate_global_output(self, model_outputs, normalized=True):
        """
        Calculates the (normalized) global output of the model.
        :param model_outputs: An instance containing node weights defined in the ModelOutputs class.
        :param normalized: Whether to normalize the global output.
        :return: float
        """
        if not self._attractors:
            self._global_output = None
            logging.debug('No attractors were found')
            return self._global_output

        pred_global_output = 0.0

        for attractor in self._attractors:
            for node_name, node_weight in model_outputs.model_outputs.items():
                if node_name not in attractor:
                    continue
                node_state = attractor[node_name]
                state_value = int(node_state) if node_state in [0, 1] else 0.5
                pred_global_output += state_value * node_weight

        pred_global_output /= len(self._attractors)
        if normalized:
            self._global_output = (pred_global_output - model_outputs.min_output) / (
                    model_outputs.max_output - model_outputs.min_output)
        else:
            self._global_output = pred_global_output
        return self._global_output

    def from_binary(self, binary_representation, mutation_type: str):
        """
        Updates the Boolean Equations from a binary representation based on the mutation type.
        :param binary_representation: The binary representation of the Boolean Equations as a list.
        :param mutation_type: The type of mutation can be: 'topology', 'balanced', 'mixed'.
        :return: Updated Boolean Equations as a list.
        """
        index = 0
        new_link = ''

        for i, (target, activating, inhibitory, link) in enumerate(self._boolean_equations):
            num_activating = len(activating)
            num_inhibitory = len(inhibitory)

            if mutation_type in ['topology', 'mixed'] and len(activating) + len(inhibitory) <= 1:
                continue

            elif mutation_type in ['topology', 'mixed']:
                new_activating_values = binary_representation[index:index + num_activating]
                index += num_activating
                new_inhibitory_values = binary_representation[index:index + num_inhibitory]
                index += num_inhibitory

                if num_activating > 0 and num_inhibitory > 0:
                    if (all(val == 0 for val in new_activating_values) and
                            all(val == 0 for val in new_inhibitory_values)):
                        new_activating_values[0] = 1

                elif num_activating > 0 and num_inhibitory == 0:
                    if all(val == 0 for val in new_activating_values):
                        new_activating_values[0] = 1

                elif num_inhibitory > 0 and num_activating == 0:
                    if all(val == 0 for val in new_inhibitory_values):
                        new_inhibitory_values[0] = 1

                new_activating = dict(zip(activating.keys(), new_activating_values))
                new_inhibitory = dict(zip(inhibitory.keys(), new_inhibitory_values))

                if mutation_type == 'mixed' and link != '':
                    link_value = binary_representation[index]
                    index += 1
                    link = '&' if link_value == 1 else '|'

                self._boolean_equations[i] = (target, new_activating, new_inhibitory, link)

            elif mutation_type == 'balanced':
                if link != '':
                    link_value = binary_representation[index]
                    index += 1
                    link = '&' if link_value == 1 else '|'
                    self._boolean_equations[i] = (target, activating, inhibitory, link)

        return self._boolean_equations

    def to_binary(self, mutation_type: str):
        """
        Converts the Boolean Equations to a binary representation. The representation is based on the mutation type.
        :param mutation_type: The type of mutation can be: 'topology', 'balanced', 'mixed'
        :return: The binary representation of the Boolean Equations as a list.
        """
        binary_representation = []

        for _, activating, inhibitory, link in self._boolean_equations:
            if mutation_type in ['topology', 'mixed'] and len(activating) + len(inhibitory) > 1:
                binary_representation.extend(int(val) for val in activating.values())
                binary_representation.extend(int(val) for val in inhibitory.values())
                if mutation_type == 'mixed' and link != '':
                    binary_representation.append(1 if link == '&' else 0)

            elif mutation_type == 'balanced' and link != '':
                binary_representation.append(1 if link == '&' else 0)

        self._binary_boolean_equations = np.array(binary_representation)
        return self._binary_boolean_equations

    def add_perturbations(self, perturbations):
        """
        Adds perturbations to the Boolean Model.
        :param perturbations: A list of Perturbations.
        :return: None
        """
        for perturbation in perturbations:
            self._perturb_nodes(perturbation['targets'], perturbation['effect'])

    def print(self):
        equations = []
        link_operator_map = {'&': 'and', '|': 'or', '': ''}

        for eq in self._boolean_equations:
            target, activating, inhibitory, link = eq
            activating_nodes = [node for node, value in activating.items() if value == 1]
            inhibitory_nodes = [node for node, value in inhibitory.items() if value == 1]

            if activating_nodes and inhibitory_nodes:
                activating_part = ' or '.join(activating_nodes)
                inhibitory_part = ' or '.join(inhibitory_nodes)
                converted_link = link_operator_map.get(link, link)
                equation = f"{target} *= ({activating_part}) {converted_link} not ({inhibitory_part})"
            elif activating_nodes:
                activating_part = ' or '.join(activating_nodes)
                equation = f"{target} *= ({activating_part})"
            elif inhibitory_nodes:
                inhibitory_part = ' or '.join(inhibitory_nodes)
                equation = f"{target} *= not ({inhibitory_part})"
            else:
                equation = f"{target} *= 0"

            equations.append(equation)

        print('\n'.join(equations))

    def clone(self):
        return BooleanModel(
            model_name=self._model_name,
            attractor_tool=self._attractor_tool,
            attractor_type=self._attractor_type,
            mutation_type=self._mutation_type,
            cloned_equations=self._boolean_equations.copy(),
            cloned_binary_equations=self._binary_boolean_equations.copy()
        )

    def reset_attractors(self) -> None:
        self._attractors = []

    def has_attractors(self) -> bool:
        return bool(self._attractors)

    def has_stable_states(self) -> bool:
        return bool(self.get_stable_states())

    def has_global_output(self) -> bool:
        return bool(self.global_output)

    def get_stable_states(self) -> object:
        return [state for state in self._attractors if '*' not in state.values()]

    @property
    def mutation_type(self) -> str:
        return self._mutation_type

    @property
    def global_output(self):
        return self._global_output

    @property
    def boolean_equations(self):
        return self._boolean_equations

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def attractors(self) -> object:
        return self._attractors

    @property
    def binary_boolean_equations(self):
        return self._binary_boolean_equations

    @property
    def attractor_tool(self) -> str:
        return self._attractor_tool

    @property
    def attractor_type(self) -> str:
        return self._attractor_type

    @model_name.setter
    def model_name(self, model_name: str) -> None:
        self._model_name = model_name

    @boolean_equations.setter
    def boolean_equations(self, boolean_equations: dict) -> None:
        self._boolean_equations = boolean_equations

    @binary_boolean_equations.setter
    def binary_boolean_equations(self, binary_boolean_equations) -> None:
        self._binary_boolean_equations = binary_boolean_equations



import pandas as pd
import logging
from matplotlib import pyplot as plt
from sklearn.metrics import auc, precision_recall_curve, roc_curve


class PlotUtil:

    @staticmethod
    def plot_roc_and_pr_curve(predicted_synergy_scores, observed_synergy_scores, synergy_method, labels=None):
        """
        Plot the ROC and PR Curves for one or multiple sets of predicted synergy scores.
        :param predicted_synergy_scores: Either a list of predictions for multiple models or a single set of predictions.
        :param observed_synergy_scores: List of observed synergy scores.
        :param synergy_method: Method used for synergy scoring (for plot titles).
        :param labels: Optional list of labels for each set of predictions. If None, default labels will be generated.
        """

        single_model = not isinstance(predicted_synergy_scores[0], list)
        if single_model:
            predicted_synergy_scores = [predicted_synergy_scores]

        if labels is None:
            labels = [f"Model {i + 1}" for i in range(len(predicted_synergy_scores))]

        plt.figure(figsize=(12, 5))

        # ROC Curve
        plt.subplot(1, 2, 1)
        for idx, (predicted_scores, label) in enumerate(zip(predicted_synergy_scores, labels)):
            df = pd.DataFrame(predicted_scores, columns=['perturbation', 'synergy_score'])
            df['observed'] = df['perturbation'].apply(lambda x: 1 if x in observed_synergy_scores else 0)
            df['synergy_score'] = df['synergy_score'] * -1
            df = df.sort_values(by='synergy_score', ascending=False).reset_index(drop=True)

            logging.info(f"Predicted Data with Observed Synergies for {label}:")
            logging.info(df)

            fpr, tpr, _ = roc_curve(df['observed'], df['synergy_score'])
            roc_auc = auc(fpr, tpr)
            auc_label = f"AUC: {roc_auc:.2f}" if single_model else f"{label} AUC: {roc_auc:.2f}"
            plt.plot(fpr, tpr, lw=2, label=auc_label)

        plt.plot([0, 1], [0, 1], color='lightgrey', lw=1.2, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate (FPR)')
        plt.ylabel('True Positive Rate (TPR)')
        plt.title(f"ROC Curve, Ensemble-wise synergies ({synergy_method})")
        plt.legend(loc="lower right")
        plt.grid(lw=0.5, color='lightgrey')

        # PR Curve
        plt.subplot(1, 2, 2)
        for idx, (predicted_scores, label) in enumerate(zip(predicted_synergy_scores, labels)):
            df = pd.DataFrame(predicted_scores, columns=['perturbation', 'synergy_score'])
            df['observed'] = df['perturbation'].apply(lambda x: 1 if x in observed_synergy_scores else 0)
            df['synergy_score'] = df['synergy_score'] * -1
            df = df.sort_values(by='synergy_score', ascending=False).reset_index(drop=True)

            precision, recall, _ = precision_recall_curve(df['observed'], df['synergy_score'])
            pr_auc = auc(recall, precision)
            auc_label = f"AUC: {pr_auc:.2f}" if single_model else f"{label} AUC: {pr_auc:.2f}"
            plt.plot(recall, precision, lw=2, label=auc_label)

        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.title(f"PR Curve, Ensemble-wise synergies ({synergy_method})")
        plt.grid(lw=0.5, color='lightgrey')
        plt.plot([0, 1], [sum(df['observed']) / len(df['observed'])] * 2, linestyle='--', color='grey')
        plt.legend(loc="upper right")

        plt.tight_layout()
        plt.show()

        # Logging
        for idx, (predicted_scores, label) in enumerate(zip(predicted_synergy_scores, labels)):
            df = pd.DataFrame(predicted_scores, columns=['perturbation', 'synergy_score'])
            df['observed'] = df['perturbation'].apply(lambda x: 1 if x in observed_synergy_scores else 0)
            df['synergy_score'] = df['synergy_score'] * -1
            fpr, tpr, _ = roc_curve(df['observed'], df['synergy_score'])
            roc_auc = auc(fpr, tpr)
            precision, recall, _ = precision_recall_curve(df['observed'], df['synergy_score'])
            pr_auc = auc(recall, precision)
            logging.debug(f"ROC AUC for {label}: {roc_auc:.2f}")
            logging.debug(f"PR AUC for {label}: {pr_auc:.2f}")


    @staticmethod
    def plot_pr_curve_with_ci(pr_df, auc_pr, boot_n, plot_discrete):
        plt.figure(figsize=(8, 6))

        if plot_discrete:
            plt.step(pr_df['recall'], pr_df['precision'], color='red', label='Precision-Recall curve', where='post')
            plt.plot(pr_df['recall'], pr_df['low_precision'], linestyle='--', color='grey',
                     label='Lower Confidence Interval', marker='s')
            plt.plot(pr_df['recall'], pr_df['high_precision'], linestyle='--', color='grey',
                     label='Upper Confidence Interval', marker='s')
            for i in range(len(pr_df)):
                plt.vlines(x=pr_df['recall'].iloc[i], ymin=pr_df['low_precision'].iloc[i],
                           ymax=pr_df['high_precision'].iloc[i], alpha=0.3,
                           color='grey', linestyle='dotted')
        else:
            plt.plot(pr_df['recall'], pr_df['precision'], color='red', label='Precision-Recall curve')
            plt.fill_between(pr_df['recall'], pr_df['low_precision'], pr_df['high_precision'], color='grey',
                             alpha=0.3, label='Confidence Interval')

        plt.plot([0, 1], [0.2, 0.2], linestyle='--', color='grey')

        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.ylim(0.0, 1.1)
        plt.title(f'PR Curve with {boot_n} Samples')
        plt.legend(loc='lower left', title=f'AUC: {auc_pr:.3f}')
        plt.grid(lw=0.5, color='lightgrey')
        plt.show()



import os
import multiprocessing
import datetime
import logging
from typing import List
import pygad
import numpy as np
from numpy.random import default_rng
from joblib import Parallel, delayed
from pydruglogics import BooleanModel
from pydruglogics.input.TrainingData import TrainingData
from pydruglogics.model.BooleanModelOptimizer import BooleanModelOptimizer
from pydruglogics.utils.BNetworkUtil import BNetworkUtil


class Evolution(BooleanModelOptimizer):
    def __init__(self, boolean_model=None, training_data=None, model_outputs=None, ga_args=None, ev_args=None):
        """
        Initializes the Evolution class with a BooleanModel and genetic algorithm parameters.
        :param boolean_model: The boolean model to be evolved.
        :param training_data: Training data for the model.
        :param model_outputs: Model outputs for evaluation.
        :param ga_args: Dictionary containing all necessary arguments for pygad.
        :param ev_args: Dictionary containing all necessary arguments for running the evolution.
        """
        self._boolean_model = boolean_model
        self._mutation_type = boolean_model.mutation_type
        self._training_data = training_data or self._create_default_training_data()
        self._model_outputs = model_outputs
        self._ga_args = ga_args or {}
        self._ev_args = ev_args or {}
        self._best_boolean_models = []
        self._global_seed = self._ev_args.get('num_of_runs', 20)
        np.random.seed(self._global_seed)

        if not self._model_outputs:
            raise ValueError('Model outputs are not provided.')

    def _callback_generation(self, ga_instance):
        logging.debug(f"Generation {ga_instance.generations_completed}: Fitness values: "
                        f"{ga_instance.last_generation_fitness}")

    def _create_default_training_data(self):
        return TrainingData(observations=[(["globaloutput:1"], 1.0)])

    def _run_single_ga(self, evolution_number, initial_population):
        """
        Runs a single GA and returns the best models for this run.
        :param evolution_number: The index of the current GA run.
        :param initial_population: The initial population for the GA.
        :return: The best models for the GA.
        """
        logging.debug(f"Running GA simulation {evolution_number}...")
        ga_seed = self._global_seed + evolution_number

        ga_instance = pygad.GA(
            num_generations=self._ga_args.get('num_generations', 20),
            num_parents_mating=self._ga_args.get('num_parents_mating', 3),
            fitness_func=self.calculate_fitness,
            mutation_num_genes=self._ga_args.get('mutation_num_genes', 3),
            gene_space=[0, 1],
            gene_type=int,
            keep_elitism=self._ga_args.get('keep_elitism', 0),
            initial_population=initial_population,
            random_seed=ga_seed,
            on_generation=self._callback_generation,
            fitness_batch_size=self._ga_args.get('fitness_batch_size', 20),
            crossover_type=self._ga_args.get('crossover_type', 'single_point'),
            mutation_type=self._ga_args.get('mutation_type', 'random'),
            parent_selection_type=self._ga_args.get('parent_selection_type', 'sss'),
            stop_criteria=self._ga_args.get('stop_criteria', 'reach_99'),
            suppress_warnings=True
        )

        ga_instance.run()

        last_gen_fitness_values = np.array(ga_instance.last_generation_fitness)
        population = np.array(ga_instance.population)
        best_solutions = np.argpartition(-last_gen_fitness_values, self._ev_args.get('num_best_solutions'))[
                         :self._ev_args.get('num_best_solutions')]

        sorted_population = [(population[i], last_gen_fitness_values[i]) for i in
                             best_solutions[np.argsort(-last_gen_fitness_values[best_solutions])]]

        logging.debug(f"Best fitness in Simulation {evolution_number}: Fitness = {sorted_population[0][1]}")
        return sorted_population

    def calculate_fitness(self, ga_instance, solutions, solution_idx):
        """
        Calculates fitness for a batch of solutions. Each solution is evaluated in a batch,
        and a fitness score is returned for each.
        :param ga_instance: Instance of the GA. It is required by PayGAD.
        :param solutions: A batch of solutions (list of binary vectors). It is required by PayGAD.
        :param solution_idx: The index of the current solution batch. It is required by PayGAD.
        :return: A list of fitness values, one for each solution in the batch.
        """
        cores = self._ev_args.get('num_of_cores') if self._ev_args.get('num_of_cores') else multiprocessing.cpu_count()
        fitness_values = Parallel(n_jobs=cores, backend='loky',  max_nbytes=None)(
            delayed(self._calculate_fitness_for_solution)(solution) for solution in solutions
        )

        return [fitness * 100 for fitness in fitness_values]

    def _calculate_fitness_for_solution(self, solution):
        fitness = 0.0
        mutated_boolean_model = self._boolean_model.clone()
        mutated_boolean_model.from_binary(solution, self._mutation_type)

        for observation in self._training_data.observations:
            response = observation['response']
            weight = observation['weight']
            mutated_boolean_model.calculate_attractors(mutated_boolean_model.attractor_tool,
                                                       mutated_boolean_model.attractor_type)
            condition_fitness = 0.0

            if mutated_boolean_model.has_attractors():
                if 'globaloutput' in response[0]:
                    observed_global_output = float(response[0].split(":")[1])
                    predicted_global_output = mutated_boolean_model.calculate_global_output(self._model_outputs)
                    condition_fitness = 1.0 - abs(predicted_global_output - observed_global_output)
                    logging.debug(f"Observed Global Output: {observed_global_output:.2f}, ")
                    logging.debug(f"Predicted Global Output: {predicted_global_output:.2f}")
                else:
                    if mutated_boolean_model.has_stable_states():
                        condition_fitness += 1.0

                    total_matches = []
                    for index, attractor in enumerate(mutated_boolean_model.attractors):
                        logging.debug(f"Checking stable state no. {index + 1}")
                        match_score = 0
                        found_observations = 0
                        for node in response:
                            node_name, observed_node_state = node.split(":")
                            observed_node_state = float(observed_node_state.strip())
                            attractor_state = attractor.get(node_name, '*')
                            predicted_node_state = 0.5 if attractor_state == '*' else float(attractor_state)
                            match = 1.0 - abs(predicted_node_state - observed_node_state)
                            logging.debug(f"Match for observation on node {node_name}: {match} (1 - "
                                             f"|{predicted_node_state} - {observed_node_state}|)")
                            match_score += match
                            found_observations += 1
                        logging.debug(f"From {found_observations} observations, found {match_score} matches")
                        if found_observations > 0:
                            if mutated_boolean_model.has_stable_states():
                                condition_fitness /= (found_observations + 1)
                            else:
                                condition_fitness /= found_observations
                            match_score /= found_observations
                        total_matches.append(match_score)
                    if total_matches:
                        avg_matches = sum(total_matches) / len(total_matches)
                        logging.debug(f"Average match value through all stable states: {avg_matches}")
                        condition_fitness += avg_matches
            fitness += condition_fitness * (weight / self._training_data.weight_sum)

        return fitness

    def create_initial_population(self, population_size, num_mutations, seed=None):
        """
        Creates an initial population for the GA.
        :param population_size: The number of individuals in the population.
        :param num_mutations: The number of mutations to perform on each individual.
        :param seed: Seed for reproducibility.
        :return: List of binary vectors representing the initial population.
        """
        rng = default_rng(seed)
        initial_boolean_model = np.array(self._boolean_model.binary_boolean_equations)
        population = np.zeros((population_size, len(initial_boolean_model)), dtype=int)

        for i in range(population_size):
            population[i] = initial_boolean_model
            mutated_idx = rng.choice(len(initial_boolean_model), num_mutations, replace=False)
            population[i][mutated_idx] = 1 - population[i][mutated_idx]
        return population

    def run(self):
        cores = self._ev_args.get('num_of_cores') if self._ev_args.get('num_of_cores') else multiprocessing.cpu_count()
        num_of_runs = self._ev_args.get('num_of_runs', 20)
        initial_populations = []

        for i in range(num_of_runs):
            seed = self._global_seed + i
            initial_population = self.create_initial_population(
                population_size=self._ga_args.get('fitness_batch_size'),
                num_mutations=self._ev_args.get('num_of_init_mutation', 10),
                seed=seed
            )
            initial_populations.append((i, initial_population))

        evolution_results = Parallel(n_jobs=cores, backend='loky',  max_nbytes=None)(
            delayed(self._run_single_ga)(i, initial_population) for i, initial_population in initial_populations
        )

        self._best_boolean_models = []

        for evolution_index, models in enumerate(evolution_results, start=1):
            for solution_index, (solution, fitness) in enumerate(models, start=1):
                best_boolean_model = self._boolean_model.clone()
                best_boolean_model.updated_boolean_equations = best_boolean_model.from_binary(solution, self._mutation_type)
                best_boolean_model.binary_boolean_equations = solution
                best_boolean_model.fitness = fitness
                best_boolean_model.model_name = f"e{evolution_index}_s{solution_index}"
                self._best_boolean_models.append(best_boolean_model)

        logging.info("Training finished.")
        return self._best_boolean_models

    def save_to_file_models(self, base_folder= './results/models'):
        try:
            now = datetime.datetime.now()
            current_date = now.strftime('%Y_%m_%d')
            current_time = now.strftime('%H%M')

            subfolder = os.path.join(base_folder, f"models_{current_date}_{current_time}")
            if not os.path.exists(subfolder):
                os.makedirs(subfolder)

            for model in self._best_boolean_models:
                evolution_number = model.model_name.split("_")[0][1:]
                solution_index = model.model_name.split("_")[1][1:]
                filename = f"e{evolution_number}_s{solution_index}.bnet"
                filepath = os.path.join(subfolder, filename)

                boolean_model_bnet = f"# {current_date}, {current_time}\n"
                boolean_model_bnet += f"# Evolution: {evolution_number} Solution: {solution_index}\n"
                boolean_model_bnet += f"# Fitness Score: {model.fitness / 100.0:.3f}\n"

                boolean_equation = BNetworkUtil.to_bnet_format(model.updated_boolean_equations)
                boolean_model_bnet += boolean_equation

                with open(filepath, "w") as file:
                    file.write(boolean_model_bnet)

            logging.info(f"Models saved to {base_folder}")

        except IOError as e:
            logging.error(f"File I/O error while saving models: {str(e)}")
            raise

    @property
    def best_boolean_models(self) -> List[BooleanModel]:
        return self._best_boolean_models



import time
import logging
from typing import Any, Dict, List, Optional, Union
from pydruglogics.model.BooleanModel import BooleanModel
from pydruglogics.input.ModelOutputs import ModelOutputs
from pydruglogics.input.TrainingData import TrainingData
from pydruglogics.input.Perturbations import Perturbation
from pydruglogics.model.Evolution import Evolution
from pydruglogics.model.ModelPredictions import ModelPredictions
from pydruglogics.utils.PlotUtil import PlotUtil


def timed_execution(func):
    """
    Decorator to log the time taken by a function and provide detailed logging.
    """
    def wrapper(*args, **kwargs):
        start_time = time.time()
        logging.info(f'{func.__name__.capitalize()} started...')
        result = func(*args, **kwargs)
        logging.info(f'{func.__name__.capitalize()} completed in {time.time() - start_time:.2f} seconds.')
        return result
    return wrapper


@timed_execution
def train(boolean_model: BooleanModel, model_outputs: ModelOutputs,
          ga_args: Dict[str, Any], ev_args: Dict[str, Any], training_data: Optional[TrainingData] = None,
          save_best_models: bool = False, save_path: str = './results/models') -> List[BooleanModel]:
    """
    Train a Boolean Model using genetic algorithm and evolution strategy.
    Finds the models with the best fitness score.
    """
    evolution = Evolution(boolean_model=boolean_model, model_outputs=model_outputs,
                          training_data=training_data, ga_args=ga_args, ev_args=ev_args)
    best_boolean_models = evolution.run()
    if save_best_models:
        evolution.save_to_file_models(save_path)
    return best_boolean_models


@timed_execution
def predict(best_boolean_models: Optional[List[BooleanModel]], model_outputs: ModelOutputs,
            perturbations: Perturbation, observed_synergy_scores: List[str],
            synergy_method: str = 'bliss', run_parallel: bool = True, plot_roc_pr_curves: bool = True,
            save_predictions: bool = False, save_path: str = './results/predictions',
            model_directory: str = '', attractor_tool: str = '',
            attractor_type: str = '', cores: int = 4) -> None:
    """
    Predict model outcomes and plot the results. If best_boolean_models is None, loads models from model_directory.
    """
    if best_boolean_models is None and model_directory:
        logging.info(f"Loading models from directory: {model_directory}")
        model_predictions = ModelPredictions(
            model_outputs=model_outputs,
            perturbations=perturbations,
            model_directory=model_directory,
            attractor_tool=attractor_tool,
            attractor_type=attractor_type,
            synergy_method=synergy_method,
            boolean_models=None
        )
    else:
        model_predictions = ModelPredictions(
            boolean_models=best_boolean_models,
            perturbations=perturbations,
            model_outputs=model_outputs,
            synergy_method=synergy_method
        )

    model_predictions.run_simulations(run_parallel, cores)

    if plot_roc_pr_curves:
        PlotUtil.plot_roc_and_pr_curve(model_predictions.predicted_synergy_scores,
                                       observed_synergy_scores, synergy_method)
    if save_predictions:
        model_predictions.save_to_file_predictions(save_path)

def execute(train_params: Optional[Dict[str, Any]] = None,
            predict_params: Optional[Dict[str, Any]] = None) -> None:
    """
    Execute training and/or prediction based on the provided parameters.
    """
    start_time = time.time()
    best_boolean_models = None

    if train_params:
        best_boolean_models = train(**train_params)

    if predict_params and best_boolean_models:
        predict(best_boolean_models=best_boolean_models, **predict_params)

    if predict_params and not best_boolean_models:
        predict(best_boolean_models=None, **predict_params)

    if train_params and predict_params:
        total_time = time.time() - start_time
        logging.info(f'Total runtime for training and prediction: {total_time:.2f} seconds')



from setuptools import setup, find_packages

setup(
    name="pydruglogics",
    version="0.1.2",
    author="Laura Szekeres",
    author_email="szekereslaura98@gmail.com",
    description="PyDrugLogics: a Python package designed for constructing, optimizing Boolean models and performs in-silico perturbations of the models.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/druglogics/pydruglogics",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
    ],
    python_requires=">=3.12",
    install_requires=[
        "pygad",
        "joblib",
        "matplotlib",
        "mpbn",
        "numpy",
        "pandas",
        "scipy",
        "scikit-learn"
    ]
)


import os
import re
from typing import List
import logging


class BNetworkUtil:

    @staticmethod
    def get_file_extension(file_name: str) -> str:
        if not file_name or '.' not in file_name:
            logging.warning("No extension found in the provided file.")
            return ''

        return file_name.rsplit('.', 1)[-1]

    @staticmethod
    def remove_extension(file_ext: str) -> str:
        file_name = os.path.basename(file_ext)
        name_without_extension = file_name.rsplit('.', 1)[0]
        if name_without_extension:
            return name_without_extension
        return file_name

    @staticmethod
    def read_lines_from_file(file_name: str, skip_empty_lines_and_comments: bool = True) -> List[str]:
        lines = []
        with open(file_name, 'r') as file:
            for line in file:
                line = line.strip()
                if skip_empty_lines_and_comments:
                    if not line.startswith('#') and len(line) > 0:
                        lines.append(line)
                else:
                    lines.append(line)
        return lines

    @staticmethod
    def is_numeric_string(value: str) -> bool:
        if isinstance(value, (int, float)):
            return True
        if isinstance(value, str):
            try:
                float(value)
                return True
            except ValueError:
                return False
        return False

    @staticmethod
    def parse_interaction(interaction: str) -> dict:
        components = interaction.split()

        if len(components) != 3:
            logging.error(f"Invalid interaction format: '{interaction}'. Expected format: "
                          f"'source interaction_type target'")
            raise ValueError(f"ERROR: Wrongly formatted interaction: {interaction}")


        source, interaction_type, target = components
        interaction_map = {
            'activate': 1, 'activates': 1, '->': 1,
            'inhibit': -1, 'inhibits': -1, '-|': -1,
            '<-': 1, '|-': -1
        }

        arc = interaction_map.get(interaction_type)
        if arc is None:
            logging.error(f"Invalid interaction type '{interaction_type}' in interaction '{interaction}'")
            raise ValueError(f"ERROR: Unrecognized interaction type: {interaction_type}")

        return {
            'source': source,
            'target': target,
            'arc': arc,
            'activating_regulators': [],
            'inhibitory_regulators': []
        }

    @staticmethod
    def create_interaction(target: str) -> dict:
        return {
            'target': target,
            'activating_regulators': [],
            'inhibitory_regulators': []
        }

    @staticmethod
    def bnet_string_to_dict(bnet_string: str):
        lines = [line.strip() for line in bnet_string.split('\n') if line.strip()]
        result = {}
        for line in lines:
            node, definition = line.split(',', 1)
            node = node.strip()
            definition = definition.strip()
            result[node] = definition

        return result

    @staticmethod
    def to_bnet_format(boolean_equations):
        """
        Converts Boolean equations to the '.bnet' format.
        :param boolean_equations: Boolean equations to be converted.
        :return: Boolean Equations in BNet format.
        """
        equation_list = []

        for eq in boolean_equations:
            target, activating_regulators, inhibitory_regulators, link = eq

            target_value = f"{target}, "

            activation_terms = [regulator for regulator, value in activating_regulators.items() if value == 1]
            if activation_terms:
                activation_expression = f"({activation_terms[0]})"
                for reg in activation_terms[1:]:
                    activation_expression = f"({activation_expression} | {reg})"
            else:
                activation_expression = ''

            inhibition_terms = [regulator for regulator, value in inhibitory_regulators.items() if value == 1]
            if inhibition_terms:
                inhibition_expression = f"({inhibition_terms[0]})"
                for reg in inhibition_terms[1:]:
                    inhibition_expression = f"({inhibition_expression} | {reg})"
            else:
                inhibition_expression = ''

            if activation_expression and inhibition_expression:
                combined_expression = f"{activation_expression} {link} !{inhibition_expression}"
            elif activation_expression or inhibition_expression:
                combined_expression = activation_expression if activation_expression else f"!{inhibition_expression}"
            else:
                combined_expression = '0'

            equation_line = f"{target_value}{combined_expression}".strip()
            equation_list.append(equation_line)

        final_equation_list = '\n'.join(equation_list)
        return final_equation_list

    @staticmethod
    def create_equation_from_bnet(equation_str):
        equation = equation_str.strip()
        target, regulators = equation.split(',', 1)
        target = target.strip()
        activating_regulators = {}
        inhibitory_regulators = {}
        link = ''

        regulators = regulators.replace('(', '').replace(')', '')
        if '!' in regulators:
            parts = regulators.split('!')
            inhibitory_part = parts[1].strip()
            before_inhibitory = parts[0].strip()
            if '&' in before_inhibitory:
                link = '&'
            elif '|' in before_inhibitory:
                link = '|'

            inhibitory_nodes = re.split(r'[|&]', inhibitory_part)
            for node in inhibitory_nodes:
                node = node.strip()
                if node:
                    inhibitory_regulators[node] = 1

            activating_nodes = re.split(r'[|&]', before_inhibitory)
            for node in activating_nodes:
                node = node.strip()
                if node:
                    activating_regulators[node] = 1
        else:
            activating_nodes = re.split(r'[|&]', regulators)
            for node in activating_nodes:
                node = node.strip()
                if node:
                    activating_regulators[node] = 1

        return target, activating_regulators, inhibitory_regulators, link,
